"""Two-Team Scoreboard."""

import curses
from collections import namedtuple

import libcurses

from tf2mon.hacker import HackerAttr


class Scoreboard:
    """Two-Team Scoreboard."""

    # pylint: disable=too-many-instance-attributes

    Column = namedtuple(
        "Column",
        [
            "width",
            "th",  # format when building heading line
            "td",  # format when building detail line
            "heading",
        ],
    )

    _columns_user = [
        Column(10, "{:>10}", "{:>10}", "STEAMID"),
        Column(1, "{:1}", "{:1}", "C"),
        Column(3, "{:>3}", "{:>3}", "K"),
        Column(3, "{:>3}", "{:>3}", "D"),
        Column(4, "{:>4}", "{:4.1f}", "KD"),
        Column(3, "{:>3}", "{:>3}", "SNI"),
        Column(1, "{:1}", "{:1}", "S"),
        Column(4, "{:>4}", "{:>4}", "UID"),
        Column(25, "{:25}", "{:25}", "USERNAME"),
    ]

    _columns_steamplayer = [
        Column(20, "{:20}", "{:20}", "PERSONANAME"),
        Column(20, "{:20}", "{:20}", "REALNAME"),
        Column(1, "{:1}", "{:1}", "P"),
        Column(2, "{:2}", "{:2}", "CC"),
        Column(2, "{:2}", "{:2}", "SC"),
        Column(4, "{:4}", "{:4}", "CI"),
        Column(0, "{}", "{}", "AGE"),
    ]

    _columns = _columns_user + _columns_steamplayer

    def __init__(self, monitor, win1, color1, win2, color2):
        """Create scoreboard for two teams.

        Args:
            monitor:    the monitor.
            win1:       curses window for team 1
            color1:     base color for team 2
            win2:       curses window for team 2
            color2:     base color for team 2
        """

        # pylint: disable=too-many-arguments

        self.monitor = monitor
        self.win1 = win1
        self.color1 = color1
        self.win2 = win2
        self.color2 = color2

        self._formatted_header = None
        self._fmt_user = None
        self._fmt_steamplayer = None

        # build the format string to build the header now
        fmt = " ".join([x.th for x in self._columns])

        # build the header now
        self._formatted_header = fmt.format(*[x.heading for x in self._columns])

        # build the format string to build detail lines later
        self._fmt_user = " ".join([x.td for x in self._columns_user])

        # build the format string to build detail lines later
        self._fmt_steamplayer = " ".join([x.td for x in self._columns_steamplayer])

        # calculate each column's x-ordinate for lookup by `heading`.
        self._col_x_width_by_heading = {}
        i = 0
        for col in self._columns:
            self._col_x_width_by_heading[col.heading] = (i, col.width)
            i += col.width + 1

        # updated by `set_sort_order`
        self._sort_col_x = 0
        self._sort_col_width = 0

    def set_sort_order(self, sort_order):
        """Set sort order."""

        self._sort_col_x, self._sort_col_width = self._col_x_width_by_heading[sort_order.name]

    def show_scores(self, team1, team2):
        """Display the scoreboards."""

        libcurses.clear_mouse_handlers()

        # Fill in whatever space is left on the scoreboards with users
        # whose team is unknown; doesn't matter which side they're
        # displayed on.
        unassigned = [x for x in self.monitor.users.active_users() if not x.team]

        #
        nusers = self.win1.getmaxyx()[0] - 1
        while len(team1) < nusers and len(unassigned) > 0:
            team1.append(unassigned.pop(0))
        self._show_team_scores(self.win1, self.color1, team1)

        #
        self._show_team_scores(self.win2, self.color2, team2 + unassigned)

    def _show_team_scores(self, win, color, team):

        ncols = win.getmaxyx()[1]
        win.erase()

        # column headings
        win.addnstr(0, 0, self._formatted_header, ncols, color)

        # highlight heading of active sort column
        win.chgat(
            0, self._sort_col_x, self._sort_col_width, color | curses.A_BOLD | curses.A_ITALIC
        )

        for lineno, user in enumerate(team, start=1):
            if not user.username:
                continue
            line = self._fmt_user.format(
                user.steamid.id if user.steamid else 0,
                user.role.display,
                user.nkills,
                user.ndeaths,
                user.kdratio,
                user.nsnipes,
                user.n_status_checks,
                user.userid,
                user.username,
            )
            if user.perk:
                line += " +" + user.perk
            elif _sp := user.steamplayer:
                line += " " + self._fmt_steamplayer.format(
                    _sp.personaname
                    if (_sp.personaname and _sp.personaname != user.username)
                    else "",
                    _sp.realname or "",
                    _sp.personastate or "",
                    _sp.loccountrycode or "",
                    _sp.locstatecode or "",
                    _sp.loccityid or "",
                    _sp.age or "",
                )

            try:
                win.addnstr(lineno, 0, line, ncols, self.monitor.ui.user_color(user, color))
            except curses.error:
                break

            # register to handle mouse events on the user.
            begin_y, begin_x = win.getbegyx()
            libcurses.add_mouse_handler(
                self._onmouse, begin_y + lineno, begin_x + 0, ncols, user
            )

    def _onmouse(self, mouse: libcurses.MouseEvent, user):
        """Handle mouse events within team scoreboards."""

        if mouse.button != 1:
            return False  # not handled

        for active_user in self.monitor.users.active_users():
            active_user.selected = False
        user.selected = True
        self.monitor.ui.update_display()

        if mouse.nclicks == 2:
            user.kick(HackerAttr.CHEATER)
        elif mouse.nclicks == 3:
            user.kick(HackerAttr.RACIST)

        return True  # handled



"""A regular-expression/handler pair.

And functions to operate on a list of them.
"""

import re

from loguru import logger


class Regex:
    """A regular-expression/handler pair."""

    def __init__(self, pattern, handler=None):
        """Create object to run `handler` when `pattern` is a hit."""

        self.regex = re.compile(pattern)
        self.handler = handler

        self.match = self.regex.match
        self.search = self.regex.search

        # set by `match_list` or `search_list` to matching `re.Match(object)`.
        self.re_match_obj = None

    @staticmethod
    def match_list(string, regex_list):
        """Return the first matching `Regex` in `regex_list` using `match`.

        The object's `re_match_obj` attribute is set to the resultant `re.Match(object)`.
        """

        for regex in regex_list:
            if m := regex.regex.match(string):
                logger.log("regex", m)
                regex.re_match_obj = m
                return regex

        return None

    @staticmethod
    def search_list(string, regex_list):
        """Return the first matching `Regex` in `regex_list` using `search`.

        The object's `re_match_obj` attribute is set to the resultant `re.Match(object)`.
        """

        for regex in regex_list:
            if m := regex.regex.search(string):
                logger.log("regex", m)
                regex.re_match_obj = m
                return regex

        return None



"""Monitor Commands and their Function Keys."""

import curses
import time

import libcurses

from tf2mon.hacker import HackerAttr
from tf2mon.regex import Regex


class FKey:
    """Monitor Commands and their Function Keys."""

    # A command may be invoked by:
    #   1. keypress within game
    #   2. keypress within monitor
    #   3. commands typed into the game console
    #   4. commands typed into the monitor console

    # pylint: disable=too-few-public-methods
    # pylint: disable=too-many-instance-attributes

    cmd_prefix: str = None

    def __init__(
        self,
        cmd=None,
        game_key=None,
        curses_key=None,
        status=None,
        handler=None,
        action=None,
        doc=None,
    ):
        """Create fkey-driven monitor command.

        Create a function to be performed when key is pressed in monitor and/or in game.

        Args:
            cmd:        command name; 'HELP', 'TOGGLE-SORT', etc.

            game_key:   key to run command from within game; 'F1'.

            curses_key: key to run command from monitor console; curses.KEY_F1.

            status:     callable returning printable current value.

            handler:    action to take when game or admin key is pressed;
                        must take 1 argument: re_match_obj.

            action:     Commands with an `action` (eg cmd='KICKS-POP' has
                        action='tf2mon_kicks_pop') are bound to run their
                        actions. Commands without an `action` (eg
                        cmd='TOGGLE-SORT' has action=None) are bound to
                        echo the command name to us (eg, "echo TF2MON-
                        TOGGLE-SORT").

            doc:        Documentation.
        """

        # pylint: disable=too-many-arguments

        self.cmd = cmd
        self.game_key = game_key
        self.curses_key = curses_key
        self.status = status
        self.handler = handler
        self.doc = doc

        # configure tf2 to perform this action when key is pressed.
        self.action = action or f"echo {self.cmd_prefix}{cmd}"

        # how to recognize tf2 performing this action in the con_logfile.
        leader = (
            r"^(\d{2}/\d{2}/\d{4} - \d{2}:\d{2}:\d{2}: )?"  # anchor to head; optional timestamp
        )
        self.regex = Regex(leader + f"{self.cmd_prefix}{cmd}$", handler) if handler else None


class FKeyManager:
    """Collection of `FKey` objects."""

    def __init__(self, monitor):
        """Initialize command and function key bindings."""

        FKey.cmd_prefix = monitor.cmd_prefix
        self.monitor = monitor
        self._fkeys = self._load_fkeys()

    def create_tf2_exec_script(self, path):
        """Create tf2 exec script of keyboard BIND commands.

        Args:
            path:   tf2 exec script to create.
        """

        # write game key bindings to tf2 exec script
        with open(path, "w", encoding="utf-8") as file:
            print(f"// auto-generated {time.asctime()}", file=file)
            print(
                "\n".join(
                    [f'bind "{x.game_key}" "{x.action}"' for x in self._fkeys if x.game_key]
                ),
                file=file,
            )

    def register_curses_handlers(self):
        """Register curses key handlers."""

        for fkey in [x for x in self._fkeys if x.handler and x.curses_key]:
            libcurses.register_fkey(fkey.handler, fkey.curses_key)

    def get_regex_list(self):
        """Return list of `Regex` for all handlers."""

        return [x.regex for x in self._fkeys if x.regex]

    def get_status_line(self):
        """Return 1-line string showing current state of function keys."""

        return " ".join([x.game_key + "=" + x.status() for x in self._fkeys if x.status])

    def get_help(self) -> str:
        """Return formatted help text for man page."""

        doc = "These function keys are available in-game and in the monitor:\n"

        for fkey in [x for x in self._fkeys if x.doc]:
            doc += f"    {fkey.game_key}={fkey.doc}\n"

        return doc

    def _load_fkeys(self):
        def _on_off(k, v):
            return k.upper() if v else k

        return [
            FKey(
                cmd="HELP",
                game_key="F1",
                curses_key=curses.KEY_F1,
                status=lambda: "HELP",
                handler=lambda m: self.monitor.ui.show_help(),
            ),
            FKey(
                cmd="TOGGLE-DEBUG",
                game_key="F2",
                curses_key=curses.KEY_F2,
                status=lambda: _on_off("debug", self.monitor.ui.debug_flag.value),
                handler=lambda m: (
                    self.monitor.ui.debug_flag.toggle,
                    self.monitor.ui.show_status(),
                ),
            ),
            FKey(
                cmd="TOGGLE-TAUNT",
                game_key="F3",
                curses_key=curses.KEY_F3,
                status=lambda: _on_off("taunt", self.monitor.ui.taunt_flag.value),
                handler=lambda m: (
                    self.monitor.ui.taunt_flag.toggle,
                    self.monitor.ui.show_status(),
                ),
            ),
            FKey(
                cmd="TOGGLE-KD",
                game_key="F4",
                curses_key=curses.KEY_F4,
                status=lambda: _on_off("kd", self.monitor.ui.show_kd.value),
                handler=lambda m: (
                    self.monitor.ui.show_kd.toggle,
                    self.monitor.ui.show_status(),
                ),
            ),
            FKey(
                cmd="TOGGLE-USER-PANEL",
                game_key="F5",
                curses_key=curses.KEY_F5,
                status=lambda: self.monitor.ui.user_panel.value.name,
                handler=lambda m: (
                    self.monitor.ui.user_panel.toggle,
                    self.monitor.ui.update_display(),
                ),
            ),
            FKey(
                cmd="SWITCH-MY-TEAM",
                game_key="F6",
                curses_key=curses.KEY_F6,
                status=lambda: self.monitor.my.team.name if self.monitor.my.team else "blu",
                handler=lambda m: (
                    self.monitor.me.assign_team(self.monitor.my.opposing_team),
                    self.monitor.ui.update_display(),
                ),
            ),
            FKey(
                cmd="TOGGLE-SORT",
                game_key="F7",
                curses_key=curses.KEY_F7,
                status=lambda: self.monitor.ui.sort_order.value.name,
                handler=lambda m: (
                    self.monitor.ui.set_sort_order(self.monitor.ui.sort_order.toggle),
                    self.monitor.ui.update_display(),
                ),
            ),
            FKey(
                cmd="TOGGLE-LOG-LOCATION",
                game_key="F8",
                curses_key=curses.KEY_F8,
                status=lambda: self.monitor.ui.log_location.value.name,
                handler=lambda m: (
                    self.monitor.ui.cycle_log_location(),
                    self.monitor.ui.show_status(),
                ),
            ),
            FKey(
                cmd="SINGLE-STEP",
                game_key="KP_DEL",
                curses_key=curses.KEY_DC,
                handler=lambda m: self.monitor.admin.start_single_stepping(),
            ),
            FKey(
                cmd="KICK-LAST-CHEATER",
                game_key="[",
                curses_key=None,
                status=lambda: HackerAttr.CHEATER.name,
                handler=lambda m: self.monitor.kick_my_last_killer(HackerAttr.CHEATER),
            ),
            FKey(
                cmd="KICK-LAST-RACIST",
                game_key="]",
                curses_key=None,
                status=lambda: HackerAttr.RACIST.name,
                handler=lambda m: self.monitor.kick_my_last_killer(HackerAttr.RACIST),
            ),
            FKey(
                cmd="KICK-LAST-SUSPECT",
                game_key="\\",
                curses_key=None,
                status=lambda: HackerAttr.SUSPECT.name,
                handler=lambda m: self.monitor.kick_my_last_killer(HackerAttr.SUSPECT),
            ),
            # Numpad
            #                      kicks     admin     spams
            #                        |         |         |
            #                        v         v         v
            #                   +-----------------------------+
            #        last/      |         |         |         |
            #        latest --> |   pop   |  pull   |   pop   |
            #                   |         |         |         |
            #                   |---------+---------+---------|
            #                   |         |         |         |
            #                   |  clear  |  clear  |  clear  |
            #                   |         |  both   |         |
            #                   |---------+---------+---------|
            #        first/     |         |         |         |
            #        oldest --> | popleft |  push   | popleft |
            #                   |         |         |         |
            #                   +-----------------------------+
            # kicks
            FKey(
                cmd="KICKS-POP",
                game_key="KP_HOME",
                curses_key=curses.KEY_HOME,
                handler=lambda m: self.monitor.kicks.pop(),
                action="tf2mon_kicks_pop",
            ),
            FKey(
                cmd="KICKS-CLEAR",
                game_key="KP_LEFTARROW",
                curses_key=curses.KEY_LEFT,
                handler=lambda m: self.monitor.kicks.clear(),
                action="tf2mon_kicks_clear",
            ),
            FKey(
                cmd="KICKS-POPLEFT",
                game_key="KP_END",
                curses_key=curses.KEY_END,
                handler=lambda m: self.monitor.kicks.popleft(),
                action="tf2mon_kicks_popleft",
            ),
            # admin
            FKey(
                cmd="PULL",
                game_key="KP_UPARROW",
                curses_key=curses.KEY_UP,
                # handler=lambda m: logger.trace('pull'),
                action="tf2mon_pull",
            ),
            FKey(
                cmd="CLEAR-QUEUES",
                game_key="KP_5",
                curses_key=curses.KEY_B2,
                handler=lambda m: self.monitor.msgqueues.clear(),
                action="tf2mon_clear_queues",
            ),
            FKey(
                cmd="PUSH",
                game_key="KP_DOWNARROW",
                curses_key=curses.KEY_DOWN,
                # handler=lambda m: logger.trace('push'),
                action="tf2mon_push",
            ),
            # spams
            FKey(
                cmd="SPAMS-POP",
                game_key="KP_PGUP",
                curses_key=curses.KEY_PPAGE,
                handler=lambda m: self.monitor.spams.pop(),
                action="tf2mon_spams_pop",
            ),
            FKey(
                cmd="SPAMS-CLEAR",
                game_key="KP_RIGHTARROW",
                curses_key=curses.KEY_RIGHT,
                handler=lambda m: self.monitor.spams.clear(),
                action="tf2mon_spams_clear",
            ),
            FKey(
                cmd="SPAMS-POPLEFT",
                game_key="KP_PGDN",
                curses_key=curses.KEY_NPAGE,
                handler=lambda m: self.monitor.spams.popleft(),
                action="tf2mon_spams_popleft",
            ),
        ]



"""Command line interface."""

import os
import sys
from pathlib import Path
from typing import List, Optional

import steam.steamid
import xdg
from libcli import BaseCLI
from loguru import logger

from tf2mon.conlog import Conlog
from tf2mon.logger import configure_logger
from tf2mon.tf2monitor import TF2Monitor


class Tf2monCLI(BaseCLI):
    """Command line interface."""

    _cachedir = xdg.xdg_cache_home() / __package__

    config = {
        # name of config file.
        "config-file": Path.home().joinpath(".tf2mon.toml"),
        #
        # toml [section-name].
        "config-name": "tf2mon",
        #
        # location of game to monitor
        "tf2_install_dir": Path.home().joinpath(
            "SteamLibrary", "steamapps", "common", "Team Fortress 2", "tf"
        ),
        #
        # basename relative to `tf2_install_dir`.
        "con_logfile": Path("console.log"),
        #
        # application name.
        "app-name": "TF2MON",
        # databases.
        "players": _cachedir / "steamplayers.db",
        "hackers-base": _cachedir / "playerlist.milenko-list.json",
        "hackers-local": _cachedir / "playerlist.tf2mon-list.json",
        # this player.
        "player_name": "Bad Dad",
    }

    exclude_print_config = [
        "app-name",
    ]

    def init_parser(self) -> None:
        """Initialize argument parser."""

        self.parser = self.ArgumentParser(
            prog=__package__,
            description=self.dedent(
                """
    Team Fortress II (`TF2`) Console Monitor (`%(prog)s`) is an interactive
    terminal application that displays real-time game state and player
    statistics of an actively running TF2 game.

    `%(prog)s` recognizes and tracks `cathook-bots`, and provides both
    application and in-game key-bindings to quickly `CALLVOTE KICK`
    hackers, mark players as racist, suspect, etc.

    `%(prog)s` reads TF2's console logfile (`con_logfile`), creates action
    scripts in TF2's `cfg` directory, and binds keys for the gamer to press
    in-game to take those actions; such as issue `SAY` and `CALLVOTE KICK`
    commands.

    Other in-game key-bindings taunt gamer's last victim/killer with a
    `CHAT` message customized with their name, k/d ratio, weapon and duel-
    score (`Taunt` and `Throe`).

    By default, `%(prog)s` starts reading the con_logfile from its end
    (`--no-rewind`), and continues to `--follow` it until interrupted.
                """
            ),
        )

    def add_arguments(self) -> None:
        """Add arguments to parser."""

        rewind = self.parser.add_mutually_exclusive_group()
        arg = rewind.add_argument(
            "--rewind",
            dest="rewind",
            action="store_true",
            default=False,
            help="start from head of logfile",
        )
        self.add_default_to_help(arg)

        arg = rewind.add_argument(
            "--no-rewind",
            dest="rewind",
            action="store_false",
            help="start from tail of logfile",
        )
        self.add_default_to_help(arg)

        follow = self.parser.add_mutually_exclusive_group()
        arg = follow.add_argument(
            "--follow",
            dest="follow",
            action="store_true",
            default=True,
            help="follow end of logfile forever",
        )
        self.add_default_to_help(arg)

        arg = follow.add_argument(
            "--no-follow",
            dest="follow",
            action="store_false",
            help="exit at end of logfile",
        )
        self.add_default_to_help(arg)

        arg = self.parser.add_argument(
            "--tf2-install-dir",
            metavar="DIR",
            default=Path(self.config["tf2_install_dir"]),
            type=Path,
            help="TF2 installation directory",
        )
        self.add_default_to_help(arg)

        arg = self.parser.add_argument(
            "con_logfile",
            default=Path(self.config["con_logfile"]),
            nargs="?",
            type=Path,
            help="TF2 console logfile; relative to `--tf2-install-dir`",
        )
        self.add_default_to_help(arg)

        self.parser.add_argument(
            "--list-con-logfile",
            action="store_true",
            help="show path to logfile and exit",
        )

        self.parser.add_argument(
            "--trunc-con-logfile",
            action="store_true",
            help="truncate logfile and exit",
        )

        self.parser.add_argument(
            "--clean-con-logfile",
            action="store_true",
            help="filter-out excluded lines from logfile to stdout and exit",
        )

        group = self.parser.add_argument_group("Debugging options")

        group.add_argument(
            "--single-step",
            action="store_true",
            help="single-step at startup",
        )

        group.add_argument(
            "--break",
            dest="breakpoint",
            type=int,
            metavar="LINENO",
            help="single-step at line `LINENO`",
        )

        group.add_argument(
            "--search",
            metavar="PATTERN",
            help="single-step when line matches `PATTERN`; add `/i` to ignore case.",
        )

        group.add_argument(
            "--inject-cmd",
            dest="inject_cmds",
            metavar="LINENO:CMD",
            action="append",
            help="inject `CMD` before line `LINENO`",
        )

        group.add_argument(
            "--inject-file",
            metavar="FILE",
            help="read list of inject commands from `FILE`",
        )

        group = self.parser.add_argument_group("Database options")

        arg = group.add_argument(
            "--players",
            metavar="FILE",
            default=Path(self.config["players"]),
            type=Path,
            help="cache `steam` user data" "",
        )
        self.add_default_to_help(arg)

        arg = group.add_argument(
            "--hackers-base",
            metavar="FILE",
            default=Path(self.config["hackers-base"]),
            type=Path,
            help="upstream hackers database",
        )
        self.add_default_to_help(arg)

        arg = group.add_argument(
            "--hackers-local",
            metavar="FILE",
            default=Path(self.config["hackers-local"]),
            type=Path,
            help="local hackers database",
        )
        self.add_default_to_help(arg)

        group.add_argument(
            "--print-steamid",
            metavar="STEAMID",
            help="print `ISteamUser.GetPlayerSummaries` for `STEAMID` and exit" "",
        )

        self.parser.add_argument_group(
            "Configuration file",
            self.dedent(
                """
    The configuration file (see `--config FILE` below) defines local
    settings:

        [%(prog)s]
        tf2_install_dir = "/path/to/your/tf2/installation"
        webapi_key = "your-steamworks-webapi-key"
        player_name = "Your Name"
                """
            ),
        )

        self.parser.add_argument_group(
            "In-Game Controls, Numpad",
            self.dedent(
                """
    While playing TF2, use the `Numpad` to kick cheaters, and generate
    `Taunt` and death-`Throe` spam.

    Messages are placed into queues, and may be popped off either end.

        Queues -->    Kicks     Admin     Spams
                        |         |         |
                        v         v         v
                   +-----------------------------+
        last/      |         |         |         |
        newest --> |   pop   |  pull   |   pop   |
                   |         |         |         |
                   |---------+---------+---------|
                   |         |         |         |
                   |  clear  |  clear  |  clear  |
                   |         |  both   |         |
                   |---------+---------+---------|
        first/     |         |         |         |
        oldest --> | popleft |  push   | popleft |
                   |         |         |         |
                   +-----------------------------+

    To vet new players that have joined the game, and to remove inactive
    players, press `NUMPAD-DOWNARROW`. The monitor indicates when new
    players have arrived and that key should be pressed; or press it each
    time you die.

    When detected, `%(prog)s` pushes hackers onto the `Kicks` queue, and
    alerts the gamer, who may then press `HOME`/`END` to issue `CHAT`
    and `CALLVOTE KICK` commands.

    When gamer kills an opponent, `%(prog)s` pushes a `Taunt` onto the
    `Spams` queue; on death, a `Throe`. Enable/disable with `F3`. Send
    with `PGUP`/`PGDN`.

    The monitor can only push actions onto the queues; gamer must pop
    for action to be taken, or clear to discard.
                """
            ),
        )

        self.parser.add_argument_group(
            "Duels",
            self.dedent(
                """
    The user-panel displays battles with opponents, grouped by weapon (and
    its state when fired).

        ┌─────────────────────────────────────────────────────────────────────┐
        │Duels:                                                               │
        │ 4 and  2 vs GLaDOS                                                  │
        │             K  2 soldier        quake_rl                            │
        │             K  1 pyro           flamethrower                        │
        │             K  1 pyro     +crit flamethrower         +Low Gravity   │
        │             D  2 heavy          minigun                             │
        │ 5 and  0 vs CreditToTeam                                            │
        │             K  2 soldier        quake_rl                            │
        │             K  2 pyro     +crit flamethrower                        │
        │             K  1 pyro           flamethrower                        │
        │ 3 and  1 vs Aperture Science Prototype XR7                          │
        │             K  1 soldier        world                               │
        │             K  1 soldier        quake_rl             +Invisibility  │
        │             K  1 pyro     +crit flamethrower                        │
        │             D  1 demo           player                              │
        └─────────────────────────────────────────────────────────────────────┘
                """
            ),
        )

        self.parser.add_argument_group(
            "Function Keys",
            self.dedent(
                """
    These function keys are available in-game and in the monitor:

        F1 Display help.
        F2 Control `say` vs `echo`.
        F3 Enable taunts and throes.
        F4 Include kd-ratio in messages.
        F5 Control User-panel display: Kicks, Spams, Duels and Auto.
        F6 Join other team.
        F7 Change scoreboard sort column.
        F8 Change logger location formats.
        [  Kick last killer as cheater.
        ]  Kick last killer as racist.
        \\  Mark last killer as suspect.
        KP_DEL Begin single-stepping.
                """
            ),
        )

        self.parser.add_argument_group(
            "Where to Operate",
            self.dedent(
                """
    `%(prog)s` works by reading the `con_logfile` to which `TF2` logs
    messages during the game. `%(prog)s` can either "tail -f" an active
    game, or `--rewind` and replay saved logfiles. Press `Enter` in the
    admin console to process the next line when in `--single-step` mode.
    Type `quit` or press `^D` to exit.

        `Duel monitors`
            Run `%(prog)s` on a secondary monitor, while playing the game on
            the primary monitor. Best performance, but a pain to `Alt-Tab`
            between monitor and game.

        `Duel machines, ssh`
            ssh from another machine and run.

        `Duel machines, nfs`
            cross-mount TF2's directory to another box and run from there.
                """
            ),
        )

        self.parser.add_argument_group(
            "Terminal Size",
            self.dedent(
                """
    `%(prog)s` requires a large terminal. Maximize the window, and use keys
    (maybe `Ctrl-Minus` and `Shift-Ctrl-Plus`) to resize. The wider the
    terminal, the more player data will be displayed:

        36x146 minimum
        42x173 display personaname
        52x211 display realname
        62x272 display age/location
                """
            ),
        )

        self.parser.add_argument_group(
            "Resizable Windows",
            self.dedent(
                """
    `Drag-and-drop` an interior border to resize the windows on either side.

    `Double-click` an interior border to enter `Resize Mode`.
        `scroll-wheel` and `arrows` move the border.
        `click`, `enter` or `esc` to exit.
                """
            ),
        )

        self.parser.add_argument_group(
            "Scoreboard",
            self.dedent(
                """
    `Single-click` user to highlight and follow.
    `Double-click` user to kick as cheater.
    `Triple-click` user to kick as racist.
    `F7` to change sort column.
                """
            ),
        )

    def main(self) -> None:
        """Command line interface entry point (method)."""

        configure_logger()

        if "webapi_key" not in self.config:
            self.parser.error("Missing config `webapi_key`")

        if self.options.con_logfile == self.config["con_logfile"]:
            # Not given on command line; prefix with effective parent.
            # (configured default is a basename relative to `tf2_install_dir`)
            self.options.con_logfile = Path(
                self.config["tf2_install_dir"],
                self.config["con_logfile"],
            )

        if self.options.single_step and not self.options.follow:
            self.parser.error("--no-follow not allowed with --single-step")

        if self.options.list_con_logfile:
            print(self.options.con_logfile)
            sys.exit(0)

        monitor = TF2Monitor(self.options, self.config)

        if self.options.trunc_con_logfile:
            Conlog(monitor).trunc()
            logger.warning(f"con_logfile {str(self.options.con_logfile)!r} truncated; Exiting.")
            sys.exit(0)

        if self.options.clean_con_logfile:
            for line in Conlog(monitor).filter_excludes():
                print(line)
            logger.warning(f"con_logfile {str(self.options.con_logfile)!r} cleaned; Exiting.")
            sys.exit(0)

        if self.options.print_steamid:
            monitor.steam_web_api.connect()
            steamid = steam.steamid.SteamID(self.options.print_steamid)
            if not steamid.is_valid():
                self.parser.error(f"invalid steamid `{self.options.print_steamid}`")
            sp = monitor.steam_web_api.find_steamid(steamid)  # noqa
            ic(sp)  # noqa
            sys.exit(0)

        if os.isatty(sys.stderr.fileno()):
            cmd = " ".join([self.parser.prog] + sys.argv[1:] + ["2>x"])
            logger.error(f"Please redirect `stderr`; try `{cmd}`")
            sys.exit(0)

        monitor.run()


def main(args: Optional[List[str]] = None) -> None:
    """Command line interface entry point (function)."""
    return Tf2monCLI(args).main()



"""Spam generator."""

from loguru import logger

from tf2mon.toggle import Cycle


class Spammer:
    """Spam generator."""

    _insults = Cycle(
        "insults",
        [
            "noob",
            "try-hard",
            "wanker",
        ],
    )

    _killed = Cycle(
        "v",
        [
            "Killed",
            # "Murdered",
            # "Slaughtered",
            # "Butchered",
            # "Exterminated",
            # "Destroyed",
            # "Annihilated",
            # "Liquidated",
            # "Decimated",
            # "Done-in",
            # "Eliminated",
            # "Eviscerated",
            # "Massacred",
            # "Snuffed-out",
        ],
    )

    _crit_taunts = Cycle(
        "ct",
        [
            # "{killed} {user} with my {weapon} +crit; {duel}",
            "{duel} vs {user}, {weapon}, +crit",
        ],
    )

    _no_crit_taunts = Cycle(
        "nct",
        [
            # "{killed} {user} with my {weapon}; {duel}",
            "{duel} vs {user}, {weapon}",
        ],
    )

    _crit_throes = Cycle(
        "cg",
        [
            # "Nice crit with that {weapon}, {user}, you {insult}!",
            "{killed} by {user} with the {weapon} +crit",
            # "{duel} vs {user}, {weapon}, +crit",
        ],
    )

    _no_crit_throes = Cycle(
        "ncg",
        [
            # "Nice {weapon} ya got there, {user}, you {insult}!",
            "{killed} by {user} with the {weapon}",
            # "{duel} vs {user}, {weapon}",
        ],
    )

    _airblast_weapons = ("world", "deflect_promode", "deflect_rocket")

    def __init__(self, monitor):
        """Initialize."""
        self.monitor = monitor

    def taunt(self, victim, weapon, crit):
        """Make noise when operator kills opponent."""

        if self.monitor.ui.taunt_flag.value:
            self._push_spam(self._crit_taunts if crit else self._no_crit_taunts, victim, weapon)

    def throe(self, killer, weapon, crit):
        """Make noise when opponent kills operator."""

        if self.monitor.ui.throe_flag.value:
            spam = self._crit_throes if crit else self._no_crit_throes
            suffix = (" +" + killer.perk) if killer.perk else ""
            self._push_spam(spam, killer, weapon, suffix)

    def _is_airblast(self, weapon):

        return weapon in self._airblast_weapons

    def _push_spam(self, messages, user, weapon, suffix=""):

        m = messages.cycle.format(
            user=user.moniker,
            duel=self.monitor.my.duel_as_str(user),
            weapon=weapon,
            killed=self._killed.cycle,
            insult=self._insults.cycle,
        )
        self.monitor.spams.push("say " + m + suffix)

    def spam(self, spamno):
        """Respond to SPAM command."""

        if spamno == 1:
            msg = str(
                "say Real-time stats brought to you by " f"{self.monitor.appname} bot detector"
            )

        else:
            logger.critical(f"bad spamno {spamno!r}")
            return

        self.monitor.spams.pushleft(msg)



"""Represent milenko's playerlist.

See https://raw.githubusercontent.com/PazerOP/tf2_bot_detector/master/schemas/v3/playerlist.schema.json
"""  # noqa

import csv
import json
import time
from collections import defaultdict
from enum import Enum
from pathlib import Path

import steam.steamid
from loguru import logger

from tf2mon.steamplayer import SteamPlayer


class HackerAttr(Enum):
    """Valid Hacker Attributes."""

    # milenko's values
    CHEATER = "cheater"
    SUSPECT = "suspect"
    EXPLOITER = "exploiter"
    RACIST = "racist"
    # tf2mon's additional values
    MILENKO = "milenko"  # track milenko's bot-hunting bots
    GAMEBOT = "bot"  # legitimate gamebot; eg `Numnutz`


class DBTYPE(Enum):
    """Valid database types."""

    BASE = "base"  # milenko's database
    LOCAL = "local"  # local addendums


class Hacker:
    """Example from milenko's playerlist.

    "players": [
        {
            "attributes": [
                "cheater"
            ],
            "last_seen": {
                "player_name": "enzic2",
                "time": 1593821733
            },
            "steamid": "[U:1:1254884]"
        },

    Hackers in tf2mon's list have additional properties:
        {

            "names": [],
        }
    """

    def __init__(
        self,
        json_player=None,
        dbtype: DBTYPE = None,
        steamid=None,
        attribute=None,
        name=None,
    ):
        """Create Hacker.

        Create Hacker either from json record, or the given values.
        The json record may be from milenko's file or a local file. Loose
        values are given when we insert a new hacker. The json record and
        the loose values are mutually exclusive, and one of them is
        required.
        """

        # pylint: disable=too-many-arguments

        if json_player:
            assert steamid is attribute is name is None
            self._from_json(json_player)
            assert dbtype in DBTYPE
            self.dbtype = dbtype
        else:
            assert steamid
            assert attribute
            assert name
            self.steamid = steamid
            self.attributes = [HackerAttr(attribute)]
            self.names = [name]
            self.last_name_seen = name
            self.last_time_seen = int(time.time())
            assert dbtype in (None, DBTYPE.LOCAL)
            self.dbtype = DBTYPE.LOCAL

    def _from_json(self, jdoc):

        # required: unique id
        self.steamid = steam.steamid.SteamID(jdoc["steamid"])

        # required: type of hacker
        self.attributes = [HackerAttr(x) for x in jdoc["attributes"]]

        # optional tf2mon extension: names this hacker has used
        self.names = jdoc.get("names", [])
        self.last_name_seen = None
        self.last_time_seen = 0

        # optional: previous appearance
        if last_seen := jdoc.get("last_seen"):
            # optional
            if name := last_seen.get("player_name"):
                if name not in self.names:
                    self.names.append(name)
                self.last_name_seen = name
            # required
            self.last_time_seen = last_seen["time"]

    @property
    def is_suspect(self):
        """Return True if hacker is suspect."""
        return HackerAttr.SUSPECT in self.attributes

    @property
    def is_racist(self):
        """Return True if hacker is racist."""
        return HackerAttr.RACIST in self.attributes

    @property
    def is_milenko(self):
        """Return True if hacker is milenko."""
        return HackerAttr.MILENKO in self.attributes

    @property
    def is_gamebot(self):
        """Return True if player is a legitimate gamebot."""
        return HackerAttr.GAMEBOT in self.attributes

    @property
    def is_banned(self):
        """Return True if hacker is suspect or racist."""
        return HackerAttr.CHEATER in self.attributes or HackerAttr.RACIST in self.attributes

    def attributes_to_str(self):
        """Return a displayable representation of the attributes."""

        return ",".join([x.value for x in self.attributes])

    def __repr__(self):
        return str(self.__dict__)

    def to_json(self):
        """Return json document for this `Hacker`."""

        jdoc = {"attributes": [x.value for x in self.attributes]}
        if self.last_time_seen:
            last_seen = jdoc["last_seen"] = {}
            if self.last_name_seen:
                last_seen["player_name"] = self.last_name_seen
            last_seen["time"] = self.last_time_seen
        jdoc["steamid"] = self.steamid.id
        if self.names:
            jdoc["names"] = self.names
        return jdoc

    def track_appearance(self, name):
        """Record user appearing as given name."""

        if name not in self.names:
            logger.warning(f"adding name {name!r}")
            self.names.append(name)

        self.last_name_seen = name
        self.last_time_seen = int(time.time())
        logger.warning(f"time {self.last_time_seen} name {self.last_name_seen!r}")


class HackerManager:
    """Collection of `Hacker`s."""

    def __init__(self, *, base: Path = None, local: Path = None):
        """Connect to `Hacker` database.

        Load records from `base`, and overlay with records from
        `local`. Save `local_path` for `save_database` to use, and
        make the records accessible via `lookup_steamid`.
        """

        self._hackers_by_steamid = {}
        if base:
            self._load(base, DBTYPE.BASE)  # load the base list
        if local:
            self._load(local, DBTYPE.LOCAL)  # merge our addendums

        # self._base = base        # we only read from this file
        self._path_local = local  # we will write to this file

        # for `lookup_name`
        self._hackers_by_name = defaultdict(list)
        for hacker in self._hackers_by_steamid.values():
            for name in hacker.names:
                self._hackers_by_name[name].append(hacker)

    def _load(self, path: Path, dbtype: DBTYPE):
        """Read playerlist at `path` loading `Hacker`s into collection."""

        logger.info(f"Reading `{path}`")
        with open(path, encoding="utf-8") as file:
            for json_player in json.load(file)["players"]:
                hacker = Hacker(json_player, dbtype)
                self._hackers_by_steamid[hacker.steamid] = hacker

    def save_database(self):
        """Save LOCAL database to disk."""

        logger.info(f"Writing {self._path_local}")
        jdoc = {
            "players": [
                x.to_json()
                for x in self._hackers_by_steamid.values()
                if x.dbtype == DBTYPE.LOCAL
            ]
        }
        with open(self._path_local, "w", encoding="utf-8") as file:
            json.dump(jdoc, file, indent=4)
            print("", file=file)

    def lookup_name(self, name):
        """Return `Hacker` for given name, else None if not found."""

        logger.trace(f"name {name}")
        return self._hackers_by_name.get(name)

    def lookup_steamid(self, steamid):
        """Return `Hacker` for given steamid, else None if not found."""

        logger.trace(f"steamid {steamid}")
        return self._hackers_by_steamid.get(steamid)

    def add(self, steamid, attribute, name):
        """Create new `Hacker`, add to LOCAL database, return new `Hacker`."""

        hacker = Hacker(steamid=steamid, attribute=attribute, name=name)
        self._hackers_by_steamid[hacker.steamid] = hacker
        self._hackers_by_name[name].append(hacker)
        return hacker

    def load_gamebots(self, path):
        """Read and create `GAMEBOT`s from list of names read from `path`.

        Used by qvalve; not used by tf2mon.
        """

        steamid = steam.steamid.SteamID(SteamPlayer.BOT_S_STEAMID)

        with open(path, encoding="utf-8") as file:
            for (name,) in csv.reader(file):
                self.add(steamid, HackerAttr.GAMEBOT, name)



"""Player 'chat' messages."""

from loguru import logger


class Chat:
    """Player 'chat' messages."""

    # pylint: disable=too-few-public-methods

    _seqno = 0

    def __init__(self, user, teamflag, msg):
        """Create `Chat` object.

        Args:
            user:       User(object).
            teamflag:   True=team-chat, False=all-chat.
            msg:        The msg chatted by the user.
        """

        self.__class__._seqno += 1
        self.seqno = self.__class__._seqno
        self.user = user
        self.teamflag = teamflag
        self.msg = msg

        level = "TEAMCHAT" if teamflag else "CHAT"
        if self.user.team:
            level += self.user.team.name

        logger.opt(depth=1).log(level, f"{user} {msg}")

    def __repr__(self):
        return (
            self.__class__.__name__
            + "("
            + ", ".join(
                [
                    f"seqno={self.seqno}",
                    f"user={self.user}",
                    f"teamflag={self.teamflag}",
                    f"msg={self.msg!r}",
                ]
            )
            + ")"
        )



"""TF2's console logfile."""

import re
import time
from collections import namedtuple

from loguru import logger


class Conlog:
    """TF2 writes console output to the file named in its `con_logfile` variable.

    Options from our command line, and commands from our admin console, may
    "inject" lines into the data stream read from the game console logfile.
    """

    # pylint: disable=too-many-instance-attributes

    _cmd = namedtuple("_cmd", ["lineno", "cmd"])

    def __init__(self, monitor) -> None:
        """Prepare to open and read the console logfile."""

        self.monitor = monitor
        self.lineno = 0
        self.is_eof = False

        self._inject_cmds = []  # list(_cmd)
        self._is_inject_sorted = False
        self._is_inject_paused = False
        self._fp = None

        #
        self.last_line = None
        self._fmt_last_line = "{lineno}: {line}"

        # ignore lines that contain these strings
        self._excludes_re = re.compile(
            "|".join(
                [
                    "Failed to find attachment point specified for",
                    "# userid name",
                    "##### CTexture::LoadTextureBitsFromFile",
                    "CMaterialVar::GetVecValue: trying to get a vec value for",
                    "Cannot update control point",
                    "DataTable warning: tf_objective_resource: Out-of-range value",
                    "EmitSound: pitch out of bounds",
                    'Error: Material "models/workshop/player/items/all_class',
                    "Lobby updated",
                    "Missing Vgui material",
                    'No such variable "',
                    "Requesting texture value from var",
                    "SOLID_VPHYSICS static prop with no vphysics model!",
                    "SetupBones: invalid bone array size",
                    "ertexlit_and_unlit_generic_bump_ps20b.vcs",
                    "m_face->glyph->bitmap.width is 0 for ch:32",
                    "to attach particle system",
                    "to get specular",
                    "unknown particle system",
                ]
            )
        )

    def open(self):
        """Wait for existence of and open console logfile."""

        while not self.monitor.options.con_logfile.exists():
            logger.warning(f"waiting for {str(self.monitor.options.con_logfile)!r}...")
            time.sleep(3)

        logger.info(f"Reading `{self.monitor.options.con_logfile}`")

        self._fp = open(  # pylint: disable=consider-using-with
            self.monitor.options.con_logfile, encoding="utf-8", errors="replace"
        )

        if not self.monitor.options.rewind:
            # would self._fp.seek(0, 2) but need lineno
            while self._fp.readline() != "":
                self.lineno += 1

            logger.log("ADMIN", f"lineno={self.lineno}")
            self.is_eof = True

    def inject_cmd_list(self, cmds):
        """Inject list of commands into logfile.

        Line numbers are found at the front of each command followed by a
        colon; (then the command).
        """

        if cmds:
            for line in cmds:
                lineno, cmd = line.strip().split(":", maxsplit=1)
                self.inject_cmd(lineno, cmd)

    def inject_cmd(self, lineno, cmd):
        """Inject command into logfile before `lineno`."""

        if not cmd.startswith(self.monitor.cmd_prefix):
            cmd = self.monitor.cmd_prefix + cmd

        self._inject_cmds.append(self._cmd(int(lineno) - 1, cmd))
        self._is_inject_sorted = False

    def readline(self):
        """Read and return next line from console lofgile.

        Return None on end-of-file, else line.strip() (which may evaluate False).
        """

        while True:

            if not self._is_inject_sorted:
                self._inject_cmds.sort(key=lambda x: x.lineno)
                self._is_inject_sorted = True

            if (
                self._inject_cmds
                and (self.is_eof or not self._is_inject_paused)
                and self._inject_cmds[0].lineno <= self.lineno
            ):

                line = self._inject_cmds.pop(0).cmd
                self._is_inject_paused = True
                self.last_line = self._fmt_last_line.format(lineno=self.lineno + 1, line=line)
                logger.log("injected", self.last_line)
                return line

            self.lineno += 1
            self._is_inject_paused = False

            try:
                line = self._fp.readline()
            except UnicodeDecodeError as err:
                logger.critical(err)
                continue

            if line:
                line = line.strip()
                if self._excludes_re.search(line):
                    logger.log("exclude", self._fmt_last_line, lineno=self.lineno, line=line)
                    continue
                self.last_line = self._fmt_last_line.format(lineno=self.lineno, line=line)
                return line

            self.is_eof = True
            if not self.monitor.options.follow:
                return None  # eof

            time.sleep(1)  # find an alternative to polling

    def trunc(self):
        """Truncate console logfile."""

        with open(self.monitor.options.con_logfile, "w", encoding="utf-8"):
            pass

    def filter_excludes(self):
        """Return list of non-excluded lines from console logfile."""

        with open(self.monitor.options.con_logfile, encoding="utf-8", errors="replace") as file:
            for line in file:
                if not self._excludes_re.match(line):
                    yield line.strip("\n")
